open Ast

let print_unaryop unary_operator =
   match unary_operator with
Negation -> print_string "!"
|Ninus -> print_string "-"


let print_oper oper =
   match oper with 
Plus -> print_string "+"
|Minus -> print_string "-"
|Times -> print_string "*"
|Divide -> print_string "/"
|Mod -> print_string "%"
|Lessthan -> print_string "<"
|Greaterthan -> print_string ">"
|Lessequal -> print_string "<="
|Greaterequal -> print_string ">="
|Equal -> print_string "=="
|Notequal -> print_string "!="
|Logicaland -> print_string "&&"
|Logicalor -> print_string "||"

let print_const constant =
	match constant with
Intc(int) -> print_int int
|Charc(char) -> let _=print_string "'" in
	              let _=print_char char in
								print_string "'"
|Floatc(float) -> print_float float

let print_constopt const_opt =
	match const_opt with
 None -> print_string "" 
|Consopt(constant) ->(*let _=print_string "=" in*) 
                     print_const constant

let print_type t = 
	match t with
|Int -> print_string "int"
|Char -> print_string "char"
|Float -> print_string "float"
|Video -> print_string "video"
|Void -> print_string "void"

let print_arraydecl arraydecl =
	match arraydecl with
Adecl(adecl_type,array_name,int)-> let _=print_type adecl_type in
                                   let _=print_string " " in
																	 let _=print_string array_name in
																	 let _=print_string "[" in
																	 let _=print_int int in
																	 print_string "]"

let print_initlist1 init =
	match init with
Init (init_name,const_opt)-> let _=print_string init_name in
                             (match const_opt with
															| None-> print_constopt const_opt
															| _-> let _=print_string "=" in
														        print_constopt const_opt)

let print_initlist2 init =
 let _=print_string ","in
 print_initlist1 init

let print_initlist i_init_list =
	match i_init_list with
hd::tl -> let _=print_initlist1 hd in
          List.iter print_initlist2 tl
| _-> print_string " "

let print_vardecl vardecl =
	match vardecl with
 Ddecl(ddecl_type,i_init_list) -> let _=print_type ddecl_type in 
                                 let _=print_string " " in
																 (*List.iter*) print_initlist i_init_list

let print_delc decl =
	match decl with
VarDecl(variable_declaration) -> let _=print_vardecl variable_declaration in
                                 print_string ";"
|ArraDecl(array_declaration) -> let _=print_arraydecl array_declaration in
                                 print_string ";"




let rec print_lv lv =
 match lv with
 Variable(variable) -> print_string variable
|Array(arrayname,expr) -> let _=print_string arrayname in
                          let _=print_string "[" in
													let _=print_expr expr in														
													print_string "]" 
|Thisl(expr) -> let _=print_string "this(" in
                let _=print_expr expr in
								print_string ")"  

and

 print_expr expr =
	match expr with
 Lval(lvalue) ->print_lv lvalue               
|Const(const_opt) -> print_constopt const_opt
|Assign(lvalue,expr) -> let _=print_lv lvalue in
                        (*(match expr with
												|Const(_)-> print_expr expr
												| _ -> let _=print_string "=" in
												       print_expr expr )*)
												let _=print_string "=" in
												print_expr expr
												
|Binop(expr1,operator,expr2) -> let _=print_expr expr1 in
                                let _=print_oper operator in
																print_expr expr2
|Unaryop(unary_operator,expr)-> let _=print_unaryop unary_operator in
                                print_expr expr
|Comma(expr1,expr2) -> let _=print_expr expr1 in
                       let _=print_string "," in
											 print_expr expr2
|This(expr1,expr2) -> let _=print_string "this(" in
                      let _=print_expr expr1 in
											let _=print_string "," in
											let _=print_expr expr2 in
											print_string ")" 
|Funcall(f_name,expr_list) -> let _=print_string f_name in
                              let _=print_string "(" in
															let _=(*List.iter*) print_expr1 expr_list in
  														print_string ")"

and 
print_expr1 expr_list =
match expr_list with
hd::tl -> let _=print_expr hd in
          List.iter print_expr3 tl
|_-> print_string " "

and
print_expr3 e =
 let _=print_string ","in
 print_expr e



let rec print_stmt stmt =
	match stmt with
 Expression(expr)-> let _=print_expr expr in
                    print_string ";\n" 
|Compound(decl_list,stmt_list)-> let _=print_string "{\n" in
	                               let _= List.iter print_delc decl_list in
                                 let _=List.iter print_stmt stmt_list in
																 print_string "\n}\n"  
|If(expr,stmt1,stmt2) -> let _= print_string "if(" in
	                     let _= print_expr expr in
											 let _= print_string ")\n" in
                       let _= print_stmt stmt1 in
											 let _= print_string "\n else\n" in
											 let _= print_stmt stmt2 in
											 print_string "\n"
											 
|While(expr,stmt) -> let _=print_string "while(" in 
	                   let _=print_expr expr in
										 let _=print_string ")\n" in
                     print_stmt stmt 

|For(expr1,expr2,expr3,stmt) ->let _=print_string "for(" in 
	                           let _=print_expr expr1 in
														 let _=print_string ";" in
                             let _=print_expr expr2 in
														 let _=print_string ";" in
														 let _=print_expr expr3 in
														 let _=print_string ")\n" in
														 print_stmt stmt 
														 
|Continue -> print_string "continue;\n"
|Break -> print_string "break;\n"
|Return(expr) -> let _=print_string "return " in 
	               let _=print_expr expr in
								 print_string ";\n"
								 
|Load(l_name,l_file,expr1,expr2)->let _=print_string "load " in 
	                               let _=print_string l_name in
																 let _=print_string " from " in
                                 let _=print_string l_file in
																 let _=print_string " with " in
																 let _=print_expr expr1 in
																 let _=print_string " and " in
																 let _=print_expr expr2 in
																 print_string ";\n"
|Store(s_name,s_file) -> let _=print_string "store " in
	                       let _=print_string s_name in
												 let _=print_string " to " in
                         let _=print_string s_file in
												 print_string ";\n"
|Insert(i_name1,i_name2,expr) -> let _=print_string "insert " in
	                               let _=print_string i_name1 in
																 let _=print_string " to " in
                                 let _=print_string i_name2 in
																 let _=print_string " from " in
																 let _=print_expr expr in
																 print_string ";\n"
|Delete(d_name,expr1,expr2) -> let _=print_string "delete " in
	                             let _=print_string d_name in
															 let _=print_string " from " in
                               let _=print_expr expr1 in
															 let _=print_string " to " in
															 let _=print_expr expr2 in
															 print_string ";\n"
|Copy(c_name1,expr1,expr2,c_name2) -> let _=print_string "copy " in
	                                    let _=print_string c_name1 in
																			let _=print_string " from " in
                                      let _=print_expr expr1 in
																			let _=print_string " to " in
																			let _=print_expr expr2 in
																			let _=print_string " to " in
																			let _=print_string c_name2 in
																			print_string ";\n"						
|Update(u_name,expr1,expr2,stmt) -> let _=print_string "update " in
	                                  let _=print_string u_name in
																		let _=print_string " from " in
                                    let _=print_expr expr1 in
																		let _=print_string " to " in
																		let _=print_expr expr2 in
																		let _=print_string "\n" in
																		print_stmt stmt
																		




let print_arg1 arg =
	match arg with
	Arg(arg_type,arg_name)-> let _= print_type arg_type in
	                         let _=print_string " " in
                           print_string arg_name
let print_arg2 arg =
 let _=print_string "," in
 print_arg1 arg

let print_arg f_arglist =
	match f_arglist with
hd::tl -> let _=print_arg1 hd in
          List.iter print_arg2 tl
| _-> print_string " "

let print_func func =
  match func with 
  Func(f_type,f_name,f_arglist,f_body) -> let _= print_type f_type in
	                                    let _=print_string " " in
                                      let _= print_string f_name in
																			let _=print_string "(" in
                                      let _=(* List.iter*) print_arg f_arglist in
																			let _=print_string ")\n" in
                                      let _=print_stmt f_body in
																			print_string "\n" 


let print_program prog =
 match prog with
 Pro(func_list) -> List.iter print_func func_list 

