open Ast

module StringMap = Map.Make(struct
		type t = string
		let compare x y = Pervasives.compare x y
	end);;

type sub = {
	is_loop : bool;
	current_type : Ast.dtype;
	is_update : bool;
	current_function_type: Ast.dtype;
	is_main:bool;
}
let type_to_string t =
	match t with
		|Int -> "int"
		|Char -> "char"
		|Float -> "float"
		|Video -> "video"
		|Void -> "void"

let get_type env =
	let (sub_env,_,_) = env in sub_env.current_type
	
let get_function env =
	let (sub_env,_,_) = env in sub_env.current_function_type
	
let get_loop env =
	let (sub_env,_,_) = env in sub_env.is_loop 
	
let set_type env t = 
	let (sub_env,v,f) = env in
	let sub_env={sub_env with current_type = t} in
	let env=(sub_env,v,f) in env  

let set_loop env t = 
	let (sub_env,v,f) = env in
	let sub_env={sub_env with is_loop = t} in 
	let env=(sub_env,v,f) in env  
let set_update env t = 
	let (sub_env,v,f) = env in
	let sub_env={sub_env with is_update = t} in 
	let env=(sub_env,v,f) in env  
 	

let check_constopt c=
	 let t = (match c with
		Intc(_)-> Int
		|Floatc(_)-> Float
		|Charc(_)->Char ) in t


let check_const env cons_opt=
	let (sub_env, var_map, func_map) = env in
	match cons_opt with
		None -> env 
		|Consopt(c) -> 
			let t = check_constopt c in
			let sub_env = {sub_env with current_type=t} in 
			let env = (sub_env, var_map, func_map) in 
	env		
	
let check_i_init env init = 
	match init with
		| Init (vid,cons_op) ->
														let (s,var_map,f) = env in
														let t = get_type env in
														let p=StringMap.mem vid var_map in 
														let _=if(p=true) then let err=  "This variable "^vid^" has been defined before\n"in
														                          print_string err  in
														let env=check_const env cons_op in
														let ctype = get_type env in
														let _=if(ctype!=t) then let err = "This variable "^vid^" has type "
														                                  ^(type_to_string t)^"but here used with"
																															^(type_to_string ctype)^"\n" in 
																															print_string err   
																															in
														let var_map = StringMap.add vid t var_map in
														let env=(s,var_map,f) in env
															
let check_vardecl env dec_list=
  match dec_list with
		Ddecl(dectype,ilist) ->
	let (sub_env,var_map,func_map) = env in
	let sub_env = {sub_env with current_type=dectype} in
	let env = (sub_env,var_map,func_map) in 
	let env = List.fold_left check_i_init env ilist in
	env
	
let check_arrdecl env dec_list=
	match dec_list with
		| Adecl(dectype,aid,length) -> 
																			let (sub_env,var_map,func_map) = env in
																			let sub_env = {sub_env with current_type=dectype} in
																			let arr_exist = StringMap.mem aid var_map in
																			let _=if(arr_exist==true) then let err =  "This array "^aid^" has been defined before\n" in
																			                                   print_string err in
																			let _=if(length<=0) then let err="This array "^aid^"'s length should be greater than 0\n" in
																			                             print_string err   in
																			let var_map = StringMap.add aid dectype var_map in
																			let env =(sub_env,var_map,func_map) in env

let check_declaration env dec_list=
	match dec_list with
		|VarDecl(var_decl) -> let env = check_vardecl env var_decl in env
		|ArraDecl(arry_decl) -> let env = check_arrdecl env arry_decl in env
	


let rec check_expression env expr =
	match expr with
		Lval (lval)-> let env = check_lval env lval in env
		|Const (c_option)-> let env = check_const env c_option in env
		|Assign (lval,expr)-> let env = check_assign env lval expr in env
		|Binop (expr1,oper,expr2)-> let env = check_binop env expr1 oper expr2 in env
		|Unaryop (uoper,expr)-> let env = check_unaryop env uoper expr in env
		|Comma (expr1,expr2)-> let env = check_comma env expr1 expr2 in env
		|This (expr1,expr2)-> let env = check_this env expr1 expr2 in env
		|Funcall (fid,expr_list)-> let env = check_funcall env fid expr_list in
	env
	
and check_unaryop env uoper expr= 
	
	let env = check_expression env expr in
	let t =get_type env in
	let _=if(t=Char) then let err= "unary operation cannot be applied to char type\n" in print_string err in
	let env = set_type env t in
	env

and check_comma env expr1 expr2=
 	
	let env = check_expression env expr1 in
	let t1 = get_type env in
	let env = check_expression env expr2 in
	let t2 = get_type env in
	let _=if(t1!=t2) then let err = "two expression in a comma expression should have the same type, but here used with "
	                                ^(type_to_string t1)^" and "^(type_to_string t2) in 
																	print_string err in
	let (sub_env,var_map,func_map) = env in
	let sub_env = {sub_env with current_type=t1} in
	let env = (sub_env,var_map,func_map) in 
	env

and check_this env expr1 expr2=
 	let (sub_env,var_map,func_map) = env in
	let env = check_expression env expr1 in
	let t1=get_type env in
	let env = check_expression env expr2 in
	let t2=get_type env in
	let _=if((t1!=Ast.Int)||(t2!=Ast.Int)) then print_string "This operation requires integer type of operands\n";   in
	let _=if(sub_env.is_update==false) then print_string " \"this\" operator should be used in update statement\n";   in
	env

and match_argu env arg expr =
	let env=check_expression env expr in
	let (sub_env,var_map,func_map) = env in
	let t1=sub_env.current_type in
	match arg with
		| Arg (t2,_) ->
										let _=if(t1!=t2) then print_string "Function's parameters are not the same\n";   in
										env

and check_funcall env fid expr_list= 
	let _ = if(fid ="main") then print_string "function main cannot be called\n" in
 	let (sub_env,var_map,func_map) = env in
	let fun_exist = StringMap.mem fid func_map in
	let _=if(fun_exist==false) then let err = "Function "^fid^" not defined\n" in print_string err  in
	let arg=StringMap.find fid func_map in
	(match arg with
		| hd::tl ->
			        
							let env=List.fold_left2 match_argu env tl expr_list in
						  (match hd with
								| Arg(t,_)->
												let env = set_type env t in 
												env)
 		|[]->env
	)
and check_binop env expr1 oper expr2=
	let (sub_env,var_map,func_map) = env in
	let env = check_expression env expr1 in
	let t1 = get_type env in
	let env = check_expression env expr2 in
	let t2 = get_type env in
	let _=if(t1!=t2) then let err="Binary operation requires same type of operands, but here used with "
	                                    ^(type_to_string t1)^" and "^(type_to_string t2)^"\n" in 
																			print_string err  in
	let _=if((t1==Ast.Float)&&(oper==Ast.Mod)) then print_string "Floating point MOD operation requires integer operands\n";   in
	let sub_env = {sub_env with current_type=t1} in
	let env = (sub_env,var_map,func_map) in 
	env

and check_assign env lval expr=
	let (sub_env, var_map, func_map) = env in
	let env = check_lval env lval in
	let t1 = get_type env in 
	let env = check_expression env expr in
	let t2 = get_type env in
	let _=if(t1!=t2) then let err="assignment requires same type of operands, but here used with"
	                              ^(type_to_string t1)^" and "^(type_to_string t1)^"\n" in print_string err   in
	let sub_env = {sub_env with current_type=t1} in
	let env = (sub_env, var_map, func_map) in 
	env
	

and check_array env aid expr=
	let (sub_env,var_map,func_map) = env in
	let env = check_expression env expr in
	let t = get_type env in
	let _= if(t!=Int) then print_string "array index should be used with type int\n";   in
	let t = StringMap.mem aid var_map in
	let _= if(t!=true) then let err="array "^aid^" not declared\n" in print_string err  in
	let vartype = StringMap.find aid var_map in
	let env = set_type env vartype in env



and check_lval env lval =
	let (_,var_map,_)=env in
	match lval with
		|Array (id,expr)-> let env = check_array env id expr in env
		|Thisl (expr)-> let env = check_expression env expr in env
    |Variable(var) -> let t = StringMap.mem var var_map in
											let _= if(t!=true) then let err ="variable "^var^" not declared" in print_string err   in
											let vartype = StringMap.find var var_map in
											let env = set_type env vartype in env



let rec check_statement env stmts =
	match stmts with 
	|Expression(expr) -> let env = check_expression env expr in env
	|Compound (decl,stmt) -> let env = check_compound env decl stmt in env
	
	(* IF: check 3 conditions:*)
	(* (1) if expression is right*)
	(* (2) if expression's type is int, we don't have bool in our language*)
	(* (3) check the two statements if right *)
	|If (expr,stmt1,stmt2) -> (match expr with
												|Binop(_,op,_) -> (match op with 
																					|Lessthan
																					|Greaterthan
																					|Lessequal
																					|Greaterequal
																					|Equal
																					|Notequal
																					|Logicaland
																					|Logicalor -> let env = check_expression env expr in 
																											 	let (sub_env,var_map,func_map) = env in
																												let _ = if(sub_env.current_type != Int) then 
																													let err="if condition requires type of int, 
																													but here used with "^(type_to_string sub_env.current_type)
																													^"\n" in print_string err   in
																												let env = check_statement env stmt1 in
																												let env = check_statement env stmt2 in env
																					|_-> let _= print_string "if condition requires logical operation"; 
																															  in env )
													 |Unaryop(op,_) -> (match op with 
																						|Negation ->let env = check_expression env expr in 
																											 	let (sub_env,var_map,func_map) = env in
																												let _ = if(sub_env.current_type != Int) then 
																													let err="if condition requires type of int, 
																													but here used with "^(type_to_string sub_env.current_type)
																													^"\n" in print_string err   in
																												let env = check_statement env stmt1 in
																												let env = check_statement env stmt2 in env
																						|Ninus -> let _= print_string "if condition requires logical operation"; 
																															  in env )
													 |_-> let _= print_string "if condition requires logical operation"; 
																															  in env )
													
	(* While: check 3 conditions *)
	(* (1) chekc expression *)
	(* (2) chekc expression's type is int*)
	(* (3) update env as is_loop = true and check the body of loop as compound statement *)
														
	|While (expr,stmt) ->(match expr with
												|Binop(_,op,_) -> (match op with 
																					|Lessthan
																					|Greaterthan
																					|Lessequal
																					|Greaterequal
																					|Equal
																					|Notequal
																					|Logicaland
																					|Logicalor ->
																											 let env = check_expression env expr in 
																											 let (sub_env,var_map,func_map) = env in
																											 let sub_env={sub_env with is_loop = true} in
																											 let _ = if(sub_env.current_type != Int) then 
																															let err="while condition requires type of int, 
																													but here used with "^(type_to_string sub_env.current_type)
																													^"\n" in print_string err   in
																											 let env = check_statement env stmt in 
																											 let env = set_loop env false in env 
																					|_-> let _= print_string "while condition requires logical operation"; 
																															  in env )
													 |Unaryop(op,_) -> (match op with 
																						|Negation ->let env = check_expression env expr in 
																												 let (sub_env,var_map,func_map) = env in
																												 let sub_env={sub_env with is_loop = true} in
																												 let _ = if(sub_env.current_type != Int) then 
																																let err="while condition requires type of int, 
																													but here used with "^(type_to_string sub_env.current_type)
																													^"\n" in print_string err   in
																												 let env = check_statement env stmt in 
																												 let env = set_loop env false in env 
																						|Ninus -> let _= print_string "while condition requires logical operation"; 
																															  in env )
													 |_-> let _= print_string "while condition requires logical operation"; 
																															  in env )
	(* For: check conditions:*)
	(* (1) expr1 must be ass expr with int*)
	(* (2) expr2 must be binary expr with int*)
	(* (3) expr3 must be assg expr with int*)
	(* (4) update env's is_loop to be true and check statement*)
									
	|For (expr1,expr2,expr3,stmt) -> let env= (match expr1 with
																						| Assign(_) -> let env =check_expression env expr1 in env
																						| _ -> let _ = print_string "the first expression in for loop requires to be assignment";
																									   in env ) in 
																	 let t = get_type env in 
																	 let _ = if(t != Int) then 
															                print_string "for loop's first expression requires type of int, but here used with conflict";   in
																							
																						
																	 let env= (match expr2 with
																						| Binop(_,_,_) -> let env =check_expression env expr2 in env
																						| _ -> let _ = print_string "the second expression in for loop requires to be binary expression";
																									   in env ) in
																	 let t = get_type env in  
																	 let _ = if(t != Int) then 
															                print_string "for loop's second expression requires type of int, but here used with conflict";   in
																							
																							
																	 let env= (match expr3 with
																						| Assign(_) -> let env =check_expression env expr3 in env
																						| _ -> let _ = print_string "the third expression in for loop requires to be assignment";
																									   in env ) in
																	 let t = get_type env in 
																	 let _ = if(t != Int) then 
															                print_string "for loop's third expression requires type of int, but here used with conflict";   in
																	 let env = set_loop env true in
																	 let env = check_statement env stmt in
																	 let env = set_loop env false in env 
	|Continue -> let t = get_loop env in 
							 let _ = if(t != true) then print_string "continue statement must be used in loop";   in env 
	|Break ->let t = get_loop env in 
					 let _ = if(t != true) then print_string "break statement must be used in loop";   in env 
	
	(*check if return has the same type has funciton *)
	|Return (expr) -> let env = check_expression env expr in
	                  let t1 = get_type env in
										let t2 = get_function env in
										let _ = if(t1 != t2) then print_string "return statement must return the same type with its function,
										                                        but here used with conflict"; in env 
	(*Load:*)
	(*(1) if id is a declared variable*)
	(*(2) expr1 & expr2 must be int*) 									 
	|Load (var,_,expr1,expr2) -> let (sub_env,var_map,func_map) = env in
		                           let t = StringMap.mem var var_map in
											         let _ = if(t!=true) then let err= ("variable "^(var)^"is not declared") in print_string err in
															 let t = StringMap.find var var_map in
															 let _ = if(t!=Video) then print_string "variable in load must be video, but here used with conflict";  in 
															 let env = check_expression env expr1 in
															 let t = get_type env in
															 let _ = if(t != Int) then 
															                print_string " expression used in load requires type of int, but here used with conflict";   in
															 let env = check_expression env expr2 in
															 let t = get_type env in
															 let _ = if(t != Int) then 
															                print_string " expression used in load requires type of int, but here used with conflict";   in env
																							
	|Store (var,_) -> let (sub_env,var_map,func_map) = env in
	                  let t = StringMap.mem var var_map in
										let _ = if(t!=true) then print_string  ("variable "^(var)^"is not declared");   in
										let t = StringMap.find var var_map in
										let _ = if(t!=Video) then print_string "variable in store must be video, but here used with conflict";  in env
	
	|Insert (var1,var2,expr) -> let (sub_env,var_map,func_map) = env in
															let t = StringMap.mem var1 var_map in
															let _ = if(t!=true) then print_string  ("variable "^(var1)^"is not declared");   in
															let t = StringMap.find var1 var_map in
															let _ = if(t!=Video) then print_string "variable in Insert must be video, but here used with conflict";   in 
															let t = StringMap.mem var2 var_map in
															let _ = if(t!=true) then print_string  ("variable "^(var2)^"is not declared");   in
															let t = StringMap.find var2 var_map in
															let _ = if(t!=Video) then print_string "variable in Insert must be video, but here used with conflict";  in 
															let env = check_expression env expr in
															let t = get_type env in
															let _ = if(t != Int) then 
															                print_string " expression used in insert requires type of int, but here used with conflict";   in env
														  
															
	|Delete (var,expr1,expr2) -> let (sub_env,var_map,func_map) = env in
															 let t = StringMap.mem var var_map in
															 let _ = if(t!=true) then print_string  ("variable "^(var)^"is not declared");   in
															 let t = StringMap.find var var_map in
															 let _ = if(t!=Video) then print_string "variable in Delete must be video, but here used with conflict";  in
															 let env = check_expression env expr1 in
															 let t = get_type env in
														   let _ = if(t != Int) then 
															                print_string " expression used in Delete requires type of int, but here used with conflict";  in 
															 let env = check_expression env expr2 in
															 let t = get_type env in
														   let _ = if(t != Int) then 
															                print_string " expression used in Delete requires type of int, but here used with conflict";   in env				
																							 					 
	|Copy (var1,expr1,expr2,var2) -> let (sub_env,var_map,func_map) = env in
																	 let t = StringMap.mem var1 var_map in
																	 let _ = if(t!=true) then print_string  ("variable "^(var1)^"is not declared");   in
																	 let t = StringMap.find var1 var_map in
																	 let _ = if(t!=Video) then print_string "variable in Copy must be video, but here used with conflict";   in
																	 let t = StringMap.mem var2 var_map in
																	 let _ = if(t!=true) then print_string  ("variable "^(var2)^"is not declared");   in
																	 let t = StringMap.find var2 var_map in
																	 let _ = if(t!=Video) then print_string "variable in Copy must be video, but here used with conflict";   in
																	 let env = check_expression env expr1 in
																	 let t = get_type env in
																   let _ = if(t != Int) then 
																	                print_string " expression used in Copy requires type of int, but here used with conflict";    in 
																	 let env = check_expression env expr2 in
																	 let t = get_type env in
																   let _ = if(t != Int) then 
																	                print_string " expression used in Copy requires type of int, but here used with conflict";   in env			
																										 			
	|Update (var,expr1,expr2,stmt) -> let (sub_env,var_map,func_map) = env in
																		let env = set_update env true in
	                                  let t = StringMap.mem var var_map in
																	  let _ = if(t!=true) then print_string  ("variable "^(var)^"is not declared");   in
																	  let t = StringMap.find var var_map in
																	  let _ = if(t!=Video) then print_string "variable in Delete must be video, but here used with conflict";   in
																	  let env = check_expression env expr1 in
																	  let t = get_type env in
																    let _ = if(t != Int) then 
																	                print_string " expression used in Delete requires type of int, but here used with conflict";   in 
																	  let env = check_expression env expr2 in
																	  let t = get_type env in
																    let _ = if(t != Int) then 
																	                print_string " expression used in Delete requires type of int, but here used with conflict";    in 
																		let env = check_statement env stmt in
																		let env = set_update env false in env			
																		
and check_compound env dec_list stat_list=
	let env = List.fold_left check_declaration env dec_list in
	let env = List.fold_left check_statement env stat_list in
	env
	
let check_argument env arg = 
	match arg with
		| Arg(t,var) -> let _ =if(t==Void) then print_string ("function cannot have type \"void\" as its arguement");   in 
		                let (s,var_map,f) = env in
										let p=StringMap.mem var var_map in
										let _=if(p==true) then let err="This variable "^var^" has been defined before" in print_string err   in
										let var_map = StringMap.add var t var_map in 
										let env=(s,var_map,f) in env
										
let check_function env function_list =
	let (sub_env,var_map,func_map) = env in
	let var_map = StringMap.empty in
	let env=(sub_env,var_map,func_map) in
	match function_list with
		| Func(t, f_name, args, stmts) -> 
																			let _ = if(t=Video) then print_string ("function "^(f_name)^"cannot return video type") in 
																			let env = List.fold_left check_argument env args in
																			let (sub_env,var_map,func_map) = env in
																			let tt = StringMap.mem f_name func_map in
																			let _=if(tt=true) then print_string ("function "^(f_name)^" has been defined") in
																			let args=Arg(t,"null")::args in
																			let func_map = StringMap.add f_name args func_map in
																			let sub_env = {sub_env with current_function_type = t} in
																			if(f_name="main") then let sub_env = {sub_env with is_main=true} in
																														 let env = (sub_env,var_map,func_map) in
																														 let env = check_statement env stmts in env
																			else let env = (sub_env,var_map,func_map) in
																					 let env = check_statement env stmts in env

let check_program env p=
	match p with
		| Pro(prog) ->
  let env = List.fold_left check_function env prog in
	let (sub_env,_,_) = env in
	let _ =if(sub_env.is_main=false) then  print_string ("there is no main function in the program") in 
	p

let translate prog =
	(* Initialize the semantic analysis environment and call main program check_program *)
	let fun_map = StringMap.empty in
	let var_map = StringMap.empty in
	let sub_env = { 
		is_loop = false;
		current_type = Ast.Int; 
		is_update = false; 
		current_function_type = Void;
		is_main=false;
		}in
	let env = (sub_env,var_map,fun_map) in
	let p=check_program env prog in p
		
